/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*
 * ADIF 3.1.1
 *
 * http://www.adif.org/311/ADIF_311.htm
 */

#include <jeffpc/buffer.h>

#include "qso.h"
#include "qso_impl.h"
#include "bands.h"
#include "util.h"
#include "version.h"

#define ADIF_VERSION "3.1.1"

enum adif_type {
	AT_UUID,
	AT_DATE,
	AT_TIME,
	AT_STR,
	AT_INT,
	AT_POWER,
	AT_FREQ,
	AT_BAND,
};

struct adif_field {
	const char *name;
	const char *nvname[2];
	enum adif_type type;
	ptrdiff_t qso_offset_if_diff;
	ptrdiff_t qso_offset;
};

static const struct adif_field fields[] = {
	{ "app_hlog_uuid", { "uuid" }, AT_UUID, -1,
		offsetof(struct qso, uuid), },
	{ "qso_date", { "start" }, AT_DATE, -1,
		offsetof(struct qso, start), },
	{ "qso_date_off", { "end" }, AT_DATE, -1,
		offsetof(struct qso, end), },
	{ "time_on", { "start" }, AT_TIME, -1,
		offsetof(struct qso, start), },
	{ "time_off", { "end" }, AT_TIME, -1,
		offsetof(struct qso, end), },
	{ "comment", { "comment" }, AT_STR, -1,
		offsetof(struct qso, comment), },

	/* local / tx */
	{ "operator", { "tx", "operator" }, AT_STR, -1,
		offsetof(struct qso, tx.operator_call), },
	{ "station_callsign", { "tx", "station" }, AT_STR, -1,
		offsetof(struct qso, tx.station_call), },
	{ "tx_power", { "tx", "power" }, AT_POWER, -1,
		offsetof(struct qso, tx.power), },
	{ "freq", { "tx", "freq" }, AT_FREQ, -1,
		offsetof(struct qso, tx.freq), },
	{ "band", { }, AT_BAND, -1,
		offsetof(struct qso, tx.freq), },
	/* TODO: split mode into mode/submode */
	{ "mode", { "tx", "mode" }, AT_STR, -1,
		offsetof(struct qso, tx.mode), },
	{ "rst_sent", { "tx", "rst" }, AT_STR, -1,
		offsetof(struct qso, tx.rst), },
	{ "my_name", { "tx", "name" }, AT_STR, -1,
		offsetof(struct qso, tx.name), },
	{ "app_hlog_my_qth", { "tx", "qth" }, AT_STR, -1,
		offsetof(struct qso, tx.qth), },
	{ "my_gridsquare", { "tx", "grid" }, AT_STR, -1,
		offsetof(struct qso, tx.grid), },
	{ "my_country", { "tx", "country" }, AT_STR, -1,
		offsetof(struct qso, tx.country), },
	{ "my_state", { "tx", "state" }, AT_STR, -1,
		offsetof(struct qso, tx.state), },
	{ "my_cnty", { "tx", "county" }, AT_STR, -1,
		offsetof(struct qso, tx.county), },
	{ "app_hlog_my_cont", { "tx", "continent" }, AT_STR, -1,
		offsetof(struct qso, tx.continent), },
	{ "app_hlog_my_dxcc", { "tx", "dxcc" }, AT_INT, -1,
		offsetof(struct qso, tx.dxcc), },
	{ "my_itu_zone", { "tx", "itu" }, AT_INT, -1,
		offsetof(struct qso, tx.itu), },
	{ "my_cq_zone", { "tx", "cqz" }, AT_INT, -1,
		offsetof(struct qso, tx.cqz), },
	{ "stx", { "tx", "seq" }, AT_INT, -1,
		offsetof(struct qso, tx.seq), },

	/* remote / rx */
	{ "call", { "rx", "station" }, AT_STR, -1,
		offsetof(struct qso, rx.station_call), },
	{ "contacted_op", { "rx", "operator" }, AT_STR,
		offsetof(struct qso, rx.station_call),
		offsetof(struct qso, rx.operator_call), },
	{ "rx_power", { "rx", "power" }, AT_POWER, -1,
		offsetof(struct qso, rx.power), },
	{ "freq_rx", { "rx", "freq" }, AT_FREQ,
		offsetof(struct qso, tx.freq),
		offsetof(struct qso, rx.freq), },
	{ "band_rx", { }, AT_BAND,
		offsetof(struct qso, tx.freq),
		offsetof(struct qso, rx.freq), },
	/* TODO: split mode into mode/submode */
	{ "app_hlog_mode_rx", { "rx", "mode" }, AT_STR,
		offsetof(struct qso, tx.mode),
		offsetof(struct qso, rx.mode), },
	{ "rst_rcvd", { "rx", "rst" }, AT_STR, -1,
		offsetof(struct qso, rx.rst), },
	{ "name", { "rx", "name" }, AT_STR, -1,
		offsetof(struct qso, rx.name), },
	{ "qth", { "rx", "qth" }, AT_STR, -1,
		offsetof(struct qso, rx.qth), },
	{ "gridsquare", { "rx", "grid" }, AT_STR, -1,
		offsetof(struct qso, rx.grid), },
	{ "country", { "rx", "country" }, AT_STR, -1,
		offsetof(struct qso, rx.country), },
	{ "state", { "rx", "state" }, AT_STR, -1,
		offsetof(struct qso, rx.state), },
	{ "cnty", { "rx", "county" }, AT_STR, -1,
		offsetof(struct qso, rx.county), },
	{ "cont", { "rx", "continent" }, AT_STR, -1,
		offsetof(struct qso, rx.continent), },
	{ "dxcc", { "rx", "dxcc" }, AT_INT, -1,
		offsetof(struct qso, rx.dxcc), },
	{ "ituz", { "rx", "itu" }, AT_INT, -1,
		offsetof(struct qso, rx.itu), },
	{ "cqz", { "rx", "cqz" }, AT_INT, -1,
		offsetof(struct qso, rx.cqz), },
	{ "srx", { "rx", "seq" }, AT_INT, -1,
		offsetof(struct qso, rx.seq), },
};

static int __adif_write_cstr(struct buffer *buf, const char *name,
			     const char *value)
{
	const size_t len = value ? strlen(value) : 0;
	char lenbuf[25];
	int ret;

	if (!len)
		return 0;

	snprintf(lenbuf, sizeof(lenbuf), ":%zu>", len);

	if ((ret = buffer_append_c(buf, '<')))
		return ret;
	if ((ret = buffer_append_cstr(buf, name)))
		return ret;
	if ((ret = buffer_append_cstr(buf, lenbuf)))
		return ret;
	if (len && ((ret = buffer_append(buf, value, len))))
		return ret;
	if ((ret = buffer_append_c(buf, '\n')))
		return ret;

	return 0;
}

static int __adif_write_str(struct buffer *buf, const char *name,
			    const struct str *value)
{
	return __adif_write_cstr(buf, name, str_cstr(value));
}

static int __adif_write_uuid(struct buffer *buf, const char *name,
			     const struct xuuid *uuid)
{
	char str[XUUID_PRINTABLE_STRING_LENGTH];

	ASSERT(!xuuid_is_null(uuid));

	xuuid_unparse(uuid, str);

	return __adif_write_cstr(buf, name, str);
}

static int __adif_write_int(struct buffer *buf, const char *name,
			    struct val *value)
{
	char str[30];

	if (!value)
		return 0;

	snprintf(str, sizeof(str), "%"PRIu64, value->i);

	return __adif_write_cstr(buf, name, str);
}

static int __adif_write_float(struct buffer *buf, const char *name,
			      struct val *value, int scale, int digits)
{
	char str[30];
	uint64_t whole;
	uint64_t frac;

	if (!value)
		return 0;

	whole = value->i / scale;
	frac = value->i % scale;

	while (frac && ((frac % 10) == 0)) {
		frac /= 10;
		digits--;
	}

	if (frac)
		snprintf(str, sizeof(str), "%"PRIu64".%0*"PRIu64, whole, digits,
			 frac);
	else
		snprintf(str, sizeof(str), "%"PRIu64, whole);

	return __adif_write_cstr(buf, name, str);
}

static int __adif_write_date(struct buffer *buf, const char *name,
			     const struct time *time)
{
	char str[9];

	if (!(time->have & TIME_PART_DATE))
		return 0;

	snprintf(str, sizeof(str), "%04u%02u%02u",
		 time->year, time->month, time->day);

	return __adif_write_cstr(buf, name, str);
}

static int __adif_write_time(struct buffer *buf, const char *name,
			     const struct time *time)
{
	char str[7];

	if (!(time->have & TIME_PART_TIME_HM))
		return 0;

	if (time->have & TIME_PART_TIME_S)
		snprintf(str, sizeof(str), "%02u%02u%02u",
			 time->hour, time->min, time->sec);
	else
		snprintf(str, sizeof(str), "%02u%02u",
			 time->hour, time->min);

	return __adif_write_cstr(buf, name, str);
}

static int adif_write_additional(struct buffer *buf, const struct qso *qso)
{
	const struct nvpair *pair;
	const char *name;
	struct str *value;
	char tmp[20];
	int ret;

	if (!qso->additional)
		return 0;

	nvl_for_each(pair, qso->additional) {
		name = nvpair_name(pair);
		value = val_cast_to_str(nvpair_value(pair));

		/*
		 * TODO: The above cast may panic if the value is not a
		 * string.  This cannot happen with ADIF-imported QSOs, but
		 * may happen if the additional nvlist contains anything
		 * other than strings.
		 */

		snprintf(tmp, sizeof(tmp), ":%zu>", str_len(value));

		/* append the key: <name:len> */
		ret = buffer_append_c(buf, '<');
		if (ret)
			goto err_putref;

		ret = buffer_append_cstr(buf, name);
		if (ret)
			goto err_putref;

		ret = buffer_append_cstr(buf, tmp);
		if (ret)
			goto err_putref;

		/* append the value: blob */
		ret = buffer_append_str(buf, value);
		if (ret)
			goto err_putref;

		/* append a newline for clarity */
		ret = buffer_append_c(buf, '\n');
		if (ret)
			goto err_putref;

		str_putref(value);
	}

	return 0;

err_putref:
	str_putref(value);

	return ret;
}

int adif_write_header(struct buffer *buf)
{
	int ret;

	if ((ret = buffer_append_cstr(buf, "###\n")))
		return ret;
	if ((ret = __adif_write_cstr(buf, "adif_ver", ADIF_VERSION)))
		return ret;
	if ((ret = __adif_write_cstr(buf, "programid", "hlog")))
		return ret;
	if ((ret = __adif_write_cstr(buf, "programversion", version_string)))
		return ret;
	if ((ret = buffer_append_cstr(buf, "<eoh>\n")))
		return ret;

	return 0;
}

int adif_write_qso(struct buffer *buf, const struct qso *qso)
{
	size_t i;
	int ret;

	for (i = 0; i < ARRAY_LEN(fields); i++) {
		const struct adif_field *field = &fields[i];
		const uintptr_t base = (uintptr_t) qso;
		const void *ptr = (void *)(base + field->qso_offset);
		const void *ifdiff = (void *)(base + field->qso_offset_if_diff);
		const bool diff_only = field->qso_offset_if_diff != -1;

		ret = -EINVAL;

		switch (field->type) {
			case AT_UUID:
				ASSERT(!diff_only);
				ret = __adif_write_uuid(buf, field->name, ptr);
				break;
			case AT_DATE:
				ASSERT(!diff_only);
				ret = __adif_write_date(buf, field->name, ptr);
				break;
			case AT_TIME:
				ASSERT(!diff_only);
				ret = __adif_write_time(buf, field->name, ptr);
				break;
			case AT_STR: {
				struct str **this = (struct str **) ptr;
				struct str **other = (struct str **) ifdiff;

				if (diff_only &&
				    ((!*this && !*other) ||
				     (*this && *other &&
				      !strcmp(str_cstr(*this), str_cstr(*other))))) {
					ret = 0;
					break;
				}

				ret = __adif_write_str(buf, field->name, *this);
				break;
			}
			case AT_INT:
				ASSERT(!diff_only);
				ret = __adif_write_int(buf, field->name,
						       *(struct val **) ptr);
				break;
			case AT_POWER:
				ASSERT(!diff_only);
				ret = __adif_write_float(buf, field->name,
							 *(struct val **) ptr,
							 1000, 3);
				break;
			case AT_FREQ: {
				struct val **this = (struct val **) ptr;
				struct val **other = (struct val **) ifdiff;

				if (diff_only &&
				    ((!*this && !*other) ||
				     (*this && *other &&
				      ((*this)->i == (*other)->i)))) {
					ret = 0;
					break;
				}

				ret = __adif_write_float(buf, field->name,
							 *this, 1000000, 6);
				break;
			}
			case AT_BAND: {
				struct val **this = (struct val **) ptr;
				struct val **other = (struct val **) ifdiff;
				struct str *band;

				ret = 0;

				if (diff_only &&
				    ((!*this && !*other) ||
				     (*this && *other &&
				      ((*this)->i == (*other)->i))))
					break;

				if (!*this)
					break;

				band = bands_lookup_adif_name_by_freq((*this)->i);
				if (!band)
					break;

				ret = __adif_write_str(buf, field->name, band);

				str_putref(band);
				break;
			}
		}

		if (ret)
			return ret;
	}

	/*
	 * Write out any additional fields that aren't part of the struct
	 * qso itself.
	 */
	ret = adif_write_additional(buf, qso);
	if (ret)
		return ret;

	return buffer_append_cstr(buf, "<eor>\n");
}

static inline int get_byte(struct buffer *buf, uint8_t *byte)
{
	ssize_t ret;

	ret = buffer_read(buf, byte, sizeof(*byte));
	if (ret < 0)
		return ret;
	if (!ret)
		return -EEOF;
	return 0;
}

static inline int skip_until_le(struct buffer *buf)
{
	uint8_t byte;
	int ret;

	do {
		ret = get_byte(buf, &byte);
		if (ret)
			return ret;
	} while (byte != '<');

	return 0;
}

static int consume_header(struct buffer *input)
{
	uint8_t byte;
	char last[4];
	ssize_t ret;

	ret = get_byte(input, &byte);
	if (ret)
		return ret;

	if (byte == '<')
		return 0; /* no header */

	for (;;) {
		/* find the next field */
		ret = skip_until_le(input);
		if (ret)
			return ret;

		/* read the name */
		ret = buffer_read(input, last, sizeof(last));
		if (ret < 0)
			return ret;
		if (ret != sizeof(last))
			return -EEOF;

		/* if it is <eoh>, we're almost done */
		if (!strncmp(last, "eoh>", sizeof(last)))
			break;
	}

	/*
	 * Skip under the next '<' to leave the input in the same state
	 * regardless of if there was a header or not - with the first
	 * non-header field '<' already consumed.
	 */
	return skip_until_le(input);
}

static int nvl_set_datetime(struct nvlist *entry, const char *name,
			    const char *value, enum adif_type type)
{
	struct val **vals;
	struct val *new[6];
	struct time time;
	size_t nelem;
	int ret;

	ret = time_parse(&time, value, type == AT_DATE);
	if (ret)
		return ret;

	ret = nvl_lookup_array(entry, name, &vals, &nelem);
	if (ret)
		return ret;

	if (nelem != ARRAY_LEN(new))
		return -ERANGE;

	if (time.have & TIME_PART_DATE) {
		new[0] = VAL_ALLOC_INT(time.year);
		new[1] = VAL_ALLOC_INT(time.month);
		new[2] = VAL_ALLOC_INT(time.day);
	} else {
		new[0] = val_getref(vals[0]);
		new[1] = val_getref(vals[1]);
		new[2] = val_getref(vals[2]);
	}

	if (time.have & TIME_PART_TIME_HM) {
		new[3] = VAL_ALLOC_INT(time.hour);
		new[4] = VAL_ALLOC_INT(time.min);
	} else {
		new[3] = val_getref(vals[3]);
		new[4] = val_getref(vals[4]);
	}

	if (time.have & TIME_PART_TIME_S)
		new[5] = VAL_ALLOC_INT(time.sec);
	else
		new[5] = val_getref(vals[5]);

	return nvl_set_array_copy(entry, name, new, ARRAY_LEN(new));
}

static int __interpret_field(struct nvlist *entry, const char *const *path,
			     enum adif_type type, const char *value)
{
	struct nvlist *nvl;
	const char *name;
	int ret;

	/* no path ==> nothing to do */
	if (!path[0] && !path[1]) {
		ASSERT3S(type, ==, AT_BAND);
		return 0;
	}

	if (!strcmp(path[0], "tx") ||
	    !strcmp(path[0], "rx") ||
	    !strcmp(path[0], "addl")) {
		name = path[1];

		nvl = nvl_lookup_nvl(entry, path[0]);
		if (IS_ERR(nvl))
			return PTR_ERR(nvl);
	} else {
		ASSERT3P(path[1], ==, NULL);
		name = path[0];

		nvl = nvl_getref(entry);
	}

	switch (type) {
		case AT_UUID:
		case AT_STR:
			ret = nvl_set_cstr_dup(nvl, name, value);
			break;
		case AT_DATE:
		case AT_TIME:
			ret = nvl_set_datetime(nvl, name, value, type);
			break;
		case AT_INT: {
			uint64_t tmp;

			ret = str2u64(value, &tmp);
			if (ret)
				break;

			ret = nvl_set_int(nvl, name, tmp);
			break;
		}
		case AT_POWER: {
			int64_t milliwatts;

			ret = extract_float_as_int64(value, 1000, &milliwatts);
			if (ret)
				break;

			if (!milliwatts)
				ret = nvl_set_null(nvl, name);
			else
				ret = nvl_set_int(nvl, name, milliwatts);
			break;
		}
		case AT_FREQ: {
			int64_t freq;

			ret = extract_float_as_int64(value, 1000000, &freq);
			if (ret)
				break;

			if (!freq)
				ret = nvl_set_null(nvl, name);
			else
				ret = nvl_set_int(nvl, name, freq);
			break;
		}
		case AT_BAND:
			panic("impossible");
	}

	nvl_putref(nvl);

	return ret;
}

static int interpret_field(struct nvlist *entry,
			   struct buffer *namebuf,
			   struct buffer *valuebuf)
{
	const char *path[2] = { "addl" };
	size_t value_len;
	size_t name_len;
	char *value;
	char *name;
	size_t i;

	name = buffer_data_mutable(namebuf);
	name_len = buffer_size(namebuf) - 1;
	if (strlen(name) != name_len) {
		fprintf(stderr, "Warning: not interpreting field "
			"name with nul ('%s'; %zu != %zu)\n",
			name, strlen(name), name_len);
		return 0;
	}

	value = buffer_data_mutable(valuebuf);
	value_len = buffer_size(valuebuf) - 1;
	if (strlen(value) != value_len) {
		fprintf(stderr, "Warning: not interpreting field "
			"value with nul ('%s'; %zu != %zu)\n",
			value, strlen(value), value_len);
		return 0;
	}

	for (i = 0; i < ARRAY_LEN(fields); i++) {
		const struct adif_field *field = &fields[i];

		if (strcasecmp(name, field->name))
			continue;

		/* found */
		return __interpret_field(entry, field->nvname, field->type,
					 value);
	}

	/* stash unknown fields in struct qso's additional nvlist */

	lowercase_inplace(name);

	path[1] = name;

	return __interpret_field(entry, path, AT_STR, value);
}

int adif_parse(struct buffer *buf, struct qso *qso)
{
	struct nvlist *entry;
	struct buffer *value;
	struct buffer name;
	char rawname[64];
	uint8_t byte;
	ssize_t ret;
	size_t len;

	buffer_init_static(&name, rawname, sizeof(rawname), sizeof(rawname),
			   true);

	value = buffer_alloc(1024);
	if (IS_ERR(value))
		return PTR_ERR(value);

	/* allocate & set up basic structure */
	entry = qso_init_entry_nvl();
	if (IS_ERR(entry)) {
		ret = PTR_ERR(entry);
		goto err;
	}

	ret = consume_header(buf);
	if (ret)
		goto err;

	for (;;) {
		/* empty name & value buffers */
		VERIFY0(buffer_truncate(&name, 0));
		VERIFY0(buffer_truncate(value, 0));

		/* read field name until we get a ':' or '>' */
		for (;;) {
			ret = get_byte(buf, &byte);
			if (ret)
				goto err;

			if ((byte == ':') || (byte == '>'))
				break;

			ret = buffer_append_c(&name, byte);
			if (ret)
				goto err;
		}

		/* nul-terminate the name */
		ret = buffer_append_c(&name, '\0');
		if (ret)
			goto err;

		/* read value length until we get a ':' or '>' */
		len = 0;
		if (byte != '>') {
			for (;;) {
				ret = get_byte(buf, &byte);
				if (ret)
					goto err;

				if ((byte == ':') || (byte == '>'))
					break;

				if ((byte < '0') || (byte > '9')) {
					ret = -EINVAL;
					goto err;
				}

				len = (len * 10) + (byte - '0');
			}

			/* skip until we get a '>' */
			if (byte == ':') {
				do {
					ret = get_byte(buf, &byte);
					if (ret)
						goto err;
				} while (byte != '>');
			}
		}

		if (!strcasecmp(buffer_data(&name), "eor"))
			break;

		/* read in the value */
		ret = buffer_copy(value, buf, len);
		if (ret < 0)
			goto err;

		/* nul-terminate the value */
		ret = buffer_append_c(value, '\0');
		if (ret)
			goto err;

		/* interpret the field name-value */
		ret = interpret_field(entry, &name, value);
		if (ret)
			goto err;

		/* find the next field */
		ret = skip_until_le(buf);
		if (ret)
			return ret;
	}

	nvl_dump_file(stdout, entry);

	ret = qso_unpack_permissive(qso, entry);

err:
	if (!IS_ERR(entry))
		nvl_putref(entry);

	buffer_free(value);

	return ret;
}
