/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

#include "bands.h"
#include "qso.h"
#include "xlua.h"

struct field {
	const char *name;
	enum {
		T_PTR,
		T_CSTR_ARRAY,
		T_SUBTABLE,
	} type;
	uintptr_t offset;

	union {
		struct {
			void (*get)(lua_State *, void *);
			void (*set)(lua_State *, void **, const char *str);
		} ptr; /* T_PTR or T_CSTR_ARRAY */
		struct {
			const char *name;
			struct field *fields;
			size_t nfields;
		} subtable; /* T_SUBTABLE */
	};
};

struct lua_data {
	void *data;
	struct field *fields;
	size_t nfields;
	bool ro;
};

static void __get_str(lua_State *state, void *ptr);
static void __get_cstr_array(lua_State *state, void *ptr);
static void __get_int(lua_State *state, void *ptr);
static void __get_band_int(lua_State *state, void *ptr);
static void __set_str(lua_State *state, void **ptr, const char *str);
static void __set_int(lua_State *state, void **ptr, const char *str);

static int qso_lua_index_data(lua_State *state);
static int qso_lua_newindex_data(lua_State *state);
static int qso_lua_newindex_data_ro(lua_State *state);

struct field side_fields[] = {
	{ "mode", T_PTR, offsetof(struct qso_side, mode),
		.ptr = { __get_str, __set_str } },
	{ "freq", T_PTR, offsetof(struct qso_side, freq),
		.ptr = { __get_int, __set_int } },
	{ "band", T_PTR, offsetof(struct qso_side, freq),
		.ptr = { __get_band_int, NULL } },
	{ "rst", T_PTR, offsetof(struct qso_side, rst),
		.ptr = { __get_str, __set_str } },
	{ "name", T_PTR, offsetof(struct qso_side, name),
		.ptr = { __get_str, __set_str } },
	{ "qth", T_PTR, offsetof(struct qso_side, qth),
		.ptr = { __get_str, __set_str } },
	{ "grid", T_PTR, offsetof(struct qso_side, grid),
		.ptr = { __get_str, __set_str } },
	{ "country", T_PTR, offsetof(struct qso_side, country),
		.ptr = { __get_str, __set_str } },
	{ "state", T_PTR, offsetof(struct qso_side, state),
		.ptr = { __get_str, __set_str } },
	{ "county", T_PTR, offsetof(struct qso_side, county),
		.ptr = { __get_str, __set_str } },
	{ "continent", T_PTR, offsetof(struct qso_side, continent),
		.ptr = { __get_str, __set_str } },
	{ "dxcc", T_PTR, offsetof(struct qso_side, dxcc),
		.ptr = { __get_int, __set_int } },
	{ "itu", T_PTR, offsetof(struct qso_side, itu),
		.ptr = { __get_int, __set_int } },
	{ "cqz", T_PTR, offsetof(struct qso_side, cqz),
		.ptr = { __get_int, __set_int } },
	{ "seq", T_PTR, offsetof(struct qso_side, seq),
		.ptr = { __get_int, __set_int } },
};

struct field qso_fields[] = {
	{ "uuid", T_CSTR_ARRAY, offsetof(struct qso, uuid_str),
		.ptr = { __get_cstr_array, NULL } },
	{ "comment", T_PTR, offsetof(struct qso, comment),
		.ptr = { __get_str, __set_str } },
	{ "tx", T_SUBTABLE, offsetof(struct qso, tx),
		.subtable = { "struct side", side_fields,
			      ARRAY_LEN(side_fields) } },
	{ "rx", T_SUBTABLE, offsetof(struct qso, rx),
		.subtable = { "struct side", side_fields,
			      ARRAY_LEN(side_fields) } },
};

static void __get_str(lua_State *state, void *ptr)
{
	struct str *str = ptr;

	lua_pushstring(state, str_cstr(str));
}

static void __get_cstr_array(lua_State *state, void *ptr)
{
	lua_pushstring(state, ptr);
}

static void __get_int(lua_State *state, void *ptr)
{
	struct val *val = ptr;

	if (!val) {
		lua_pushnil(state);
	} else {
		ASSERT3S(val->type, ==, VT_INT);
		lua_pushinteger(state, val->i);
	}
}

static void __get_band_int(lua_State *state, void *ptr)
{
	struct val *val = ptr;

	if (!val) {
		lua_pushnil(state);
	} else {
		ASSERT3S(val->type, ==, VT_INT);
		lua_pushinteger(state, bands_lookup_lambda_by_freq(val->i));
	}
}

static void __set_str(lua_State *state, void **ptr, const char *str)
{
	struct str **val = (struct str **) ptr;

	str_putref(*val);
	*val = STR_DUP(str);
}

static void __set_int(lua_State *state, void **ptr, const char *str)
{
	struct val **val = (struct val **) ptr;
	uint64_t tmp;
	int ret;

	ret = str2u64(str, &tmp);
	ASSERT0(ret);

	val_putref(*val);
	*val = VAL_ALLOC_INT(tmp);
}

static inline void push_data_with_metatable(lua_State *state, void *data,
					    const char *dataname,
					    struct field *fields,
					    size_t nfields, bool ro)
{
	struct lua_data *wrapper;

	wrapper = lua_newuserdata(state, sizeof(struct lua_data));
	ASSERT3P(wrapper, !=, NULL);

	wrapper->data = data;
	wrapper->fields = fields;
	wrapper->nfields = nfields;
	wrapper->ro = ro;

	/* get the current metatable */
	if (!lua_getmetatable(state, -1))
		luaL_newmetatable(state, dataname);

	lua_pushstring(state, "__index"); /* push key */
	lua_pushcfunction(state, qso_lua_index_data); /* push value */
	lua_settable(state, -3); /* table[key] = value */

	lua_pushstring(state, "__newindex"); /* push key */
	/* push value */
	lua_pushcfunction(state,
			  ro ? qso_lua_newindex_data_ro : qso_lua_newindex_data);
	lua_settable(state, -3); /* table[key] = value */

	/* set the metatable */
	lua_setmetatable(state, -2);
}

/* handle a table[key], where table is -2 and key is -1 */
static int qso_lua_index_data(lua_State *state)
{
	struct lua_data *wrapper;
	struct lua_data data;
	const char *name;
	size_t i;

	ASSERT3U(lua_gettop(state), ==, 2);

	wrapper = lua_touserdata(state, -2);
	ASSERT3P(wrapper, !=, NULL);
	ASSERT3P(wrapper->data, !=, NULL);

	data = *wrapper;
	name = lua_tostring(state, -1);

	/* pop the arguments */
	lua_pop(state, 2);

	/* push the return values */
	for (i = 0; i < data.nfields; i++) {
		struct field *field = &data.fields[i];
		uintptr_t ptr = (uintptr_t) data.data + field->offset;

		if (strcmp(name, field->name))
			continue;

		switch (data.fields[i].type) {
			case T_PTR:
				data.fields[i].ptr.get(state, *(void **) ptr);
				break;
			case T_CSTR_ARRAY:
				data.fields[i].ptr.get(state, (void *) ptr);
				break;
			case T_SUBTABLE:
				push_data_with_metatable(state, (void *) ptr,
							 field->subtable.name,
							 field->subtable.fields,
							 field->subtable.nfields,
							 data.ro);
				break;
		}

		return 1;
	}

	return luaL_error(state, "unknown field name '%s'", name);
}

/* handle a table[key] = value, where table is -3, key is -2, and value is -1 */
static int qso_lua_newindex_data(lua_State *state)
{
	struct lua_data *wrapper;
	struct lua_data data;
	const char *name;
	size_t i;

	ASSERT3U(lua_gettop(state), ==, 3);

	wrapper = lua_touserdata(state, -3);
	ASSERT3P(wrapper, !=, NULL);
	ASSERT3P(wrapper->data, !=, NULL);

	data = *wrapper;
	name = lua_tostring(state, -2);

	for (i = 0; i < data.nfields; i++) {
		struct field *field = &data.fields[i];
		uintptr_t ptr = (uintptr_t) data.data + field->offset;

		if (strcmp(name, field->name))
			continue;

		switch (field->type) {
			case T_PTR:
				if (!field->ptr.set)
					goto readonly;

				field->ptr.set(state, (void **) ptr,
					       lua_tostring(state, -1));
				break;
			case T_CSTR_ARRAY:
				goto readonly;
			case T_SUBTABLE:
				goto readonly;
		}

		/* clean up */
		lua_pop(state, 3);

		return 0;
	}

	return luaL_error(state, "unknown field name '%s'", name);

readonly:
	return luaL_error(state, "cannot change read-only field '%s'", name);
}

/* handle a table[key] = value, when QSO is marked as read-only */
static int qso_lua_newindex_data_ro(lua_State *state)
{
	return luaL_error(state, "cannot change read-only QSO");
}

/*
 * Call a function with the qso as an argument.  If ro==true, struct qso
 * cannot be modified by the lua script.
 *
 * Returns:
 *  -ENOENT: function not found
 */
int qso_lua_call(void *state, const char *funcname, struct qso *qso, bool ro,
		 char *error, size_t errlen)
{
	int ret;

	ret = lua_getglobal(state, funcname);
	if (ret != LUA_TFUNCTION) {
		lua_pop(state, 2);
		CHECK_STACK(state);

		strcpy_safe(error, "not a function", errlen);

		return -ENOENT;
	}

	push_data_with_metatable(state, qso, "struct qso", qso_fields,
				 ARRAY_LEN(qso_fields), ro);

	/*
	 * now invoke the function
	 */
	ret = lua_pcall(state, 1, 1, 0);
	if (ret)
		COPY_ERROR(state, -1, error, errlen);

	lua_pop(state, 1);
	CHECK_STACK(state);

	return lua2errno(ret);
}
