/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

#include <jeffpc/cstr.h>

#include "xlua.h"

int lua2errno(int e)
{
	switch (e) {
		case LUA_OK:
			return 0;
		case LUA_ERRFILE: /* file-related error */
			return -EIO;
		case LUA_ERRSYNTAX: /* syntax error during precompilation */
		case LUA_ERRRUN: /* runtime error */
		case LUA_ERRERR: /* msg handler execution error */
			return -EINVAL;
		case LUA_ERRMEM: /* out of memory */
		case LUA_ERRGCMM: /* __gc metamethod error */
			return -ENOMEM;
	}

	panic("unhandled lua error %d", e);
}

void *xlua_init(const char *script_fname, char *error, size_t errlen)
{
	lua_State *state;
	int ret;

	if (!script_fname) {
		strcpy_safe(error, "invalid argument", errlen);
		return ERR_PTR(-EINVAL);
	}

	state = luaL_newstate();
	if (!state) {
		strcpy_safe(error, "failed to allocate lua state", errlen);
		return ERR_PTR(-ENOMEM);
	}

	luaL_openlibs(state);

	ret = luaL_loadfile(state, script_fname);
	if (ret)
		goto err;

	ret = lua_pcall(state, 0, 0, 0);
	if (ret)
		goto err;

	CHECK_STACK(state);

	return state;

err:
	COPY_ERROR(state, -1, error, errlen);

	lua_close(state);

	return ERR_PTR(lua2errno(ret));
}

void xlua_free(void *state)
{
	lua_close(state);
}

int xlua_call(void *state, const char *funcname, char *error, size_t errlen)
{
	int ret;

	ret = lua_getglobal(state, funcname);
	if (ret != LUA_TFUNCTION) {
		lua_pop(state, 1);
		CHECK_STACK(state);

		strcpy_safe(error, "not a function", errlen);

		return -ENOENT;
	}

	ret = lua_pcall(state, 0, 1, 0);
	if (ret)
		COPY_ERROR(state, -1, error, errlen);

	lua_pop(state, 1);
	CHECK_STACK(state);

	return lua2errno(ret);
}
